#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdint.h>
#include <string.h>
#include <sys/mman.h>
#include <errno.h>
#include <stdbool.h>
#include <stdarg.h>

#include <xengnttab.h>

#include "gup.h"

#define PAGE_SIZE 4096
#define TEST_DEVICE "/dev/xst/gup"

void error(const char *fmt, ...)
{
	va_list va;

	printf("ERROR: ");

	va_start(va, fmt);
	vprintf(fmt, va);
	va_end(va);

	if (errno)
		printf(": %s", strerror(errno));
	printf("\n");

	exit(127);
}

/**
 * Test get_user_pages() and get_user_pages_fast()
 * @buf: buffer to fill with page aligned tags.
 * @addr: address to call get_user_pages() with
 * @nr_pages: number of pages
 * @foreign: are the pages expected to be foreign?
 *
 * The driver will verify the pages get_user_pages() gets have the
 * expected tags at the start of each page.
 */
int test_gup(void *buf, void *addr, unsigned int nr_pages, bool foreign)
{
	static uint64_t magic = 0x2379F7B1398CDC00ull;
	struct xst_gup_request req;
	void *page;
	int i;
	int fd;
	int ret;

	req.address = addr;
	req.nr_pages = nr_pages;
	req.expect_magic = magic++;
	req.expect_foreign = foreign;

	for (i = 0, page = buf; i < nr_pages; i++, page += PAGE_SIZE) {
		struct xst_gup_tag *tag = page;

		tag->magic = req.expect_magic;
		tag->seq = i;
	}

	fd = open(TEST_DEVICE,  O_RDWR);
	if (fd < 0)
		error(TEST_DEVICE);

	ret = write(fd, &req, sizeof(struct xst_gup_request));

	close(fd);

	return ret == sizeof(struct xst_gup_request) ? 0 : -1;
}

static int run_local(void)
{
	const int nr_pages = 4;
	void *memory;
	int ret;

	ret = posix_memalign(&memory, PAGE_SIZE, PAGE_SIZE * nr_pages);
	if (ret < 0)
		error("posix_memalign");

	ret = test_gup(memory, memory, nr_pages, false);

	free(memory);

	return ret;
}

static int run_foreign(void)
{
	const int nr_pages = 4;
	uint32_t refs_table[nr_pages];
	void* local_map;
	void* granted_map;
	int ret;

	xengntshr_handle *gntshr;
	xengnttab_handle *gnttab;

	gntshr = xengntshr_open(NULL, 0);
	if (gntshr == NULL)
		error("xengntshr_open");

	gnttab = xengnttab_open(NULL, 0);
	if (gnttab == NULL)
		error("xengnttab_open");

	local_map = xengntshr_share_pages(gntshr, 0,
					  nr_pages, refs_table, 1);
	if (local_map == NULL)
		error("xengntshr_share_pages");

	granted_map = xengnttab_map_domain_grant_refs(gnttab, nr_pages, 0,
						      refs_table, PROT_READ | PROT_WRITE);
	if (granted_map == NULL)
		error("xengnttab_map_domain_grant_refs");

	ret = test_gup(local_map, granted_map, nr_pages, true);
	if (ret < 0)
		goto out;

	/*
	 * Using madvise() will cause a VMA split.
	 *
	 * We change the middle pages to get two splits, both
	 * new_below and !new_below (see madvise_behavior() in
	 * mm/madvise.c and __split_vma() in mm/mmap.c).
	 */
	ret = madvise(granted_map + PAGE_SIZE, 2 * PAGE_SIZE, MADV_SEQUENTIAL);
	if (ret < 0)
		error("madvise");

	ret = test_gup(local_map, granted_map, nr_pages, true);
	if (ret < 0)
		goto out;

  out:
	xengnttab_close(gnttab);
	xengntshr_close(gntshr);

	return ret;
}

int run_test(const char *name, int (*func)(void))
{
	int ret;

	ret = func();
	if (ret < 0) {
		printf("%s: FAIL\n", name);
		return 1;
	}
	printf("%s: Success\n", name);
	return 0;
}

int main(void)
{
	int fails = 0;

	fails += run_test("local", run_local);
	fails += run_test("foreign", run_foreign);

	return fails;
}
